/* $Id: pstree.c,v 1.22 2000/01/06 23:38:33 ben Exp $ */
/*
 * Copyright 1993-1999 Werner Almesberger.
 * Copyright (c) 2000 Ben Harris.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms of parts of or the
 * whole original or derived work are permitted provided that the
 * original work is properly attributed to the author. The name of the
 * author may not be used to endorse or promote products derived from
 * this software without specific prior written permission. This work
 * is provided "as is" and without any express or implied warranties.
 */
/* pstree.c - display process tree */

#include <config.h>
#include <pstree.h>

#include <sys/param.h>
#include <sys/types.h>

#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>  /* for TIOCGWINSZ */
#endif
#include <sys/stat.h>

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef HAVE_ERR_H
#include <err.h>
#endif

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef HAVE_TERMINFO
/* Note -- it's important not to have included <termios.h> here, as that
   breaks on Solaris 7 (causing curses.h to forget to include termio.h) */
#include <curses.h> /* For COLS */
#include <term.h>   /* All other terminfo stuff */
#else
#ifdef HAVE_TERMCAP
#include <termcap.h>
#endif
#endif

#ifndef MAX_DEPTH
#define MAX_DEPTH    100
#endif

#include "comm.h"

/* UTF-8 defines by Johan Myreen */
#define UTF_V	"\342\224\202\277"		/* Vertical line drawing char */
#define UTF_VR	"\342\224\234\277"		/* Vertical and right */
#define UTF_H	"\342\224\200\277"		/* Horizontal */
#define UTF_UR	"\342\224\224\277"		/* Up and right */
#define UTF_HD	"\342\224\254\277"		/* Horizontal and down */

#define VT_BEG	"\033(0\017"			/* use graphic chars */
#define VT_END	"\033(B"			/* back to normal char set */
#define VT_V	"x"				/* see UTF definitions above */
#define VT_VR	"t"
#define VT_H	"q"
#define VT_UR	"m"
#define	VT_HD	"w"

typedef struct _proc {
    char *comm;
    char **argv;	/* only used : argv[0] is 1st arg; undef if argc < 1 */
    int argc;		/* with -a   : number of arguments, -1 if swapped    */
    pid_t pid,ppid;
    uid_t uid;
    int highlight;
    struct _child *children;
    struct _proc *parent;
    struct _proc *next;
} PROC;

typedef struct _child {
    PROC *child;
    struct _child *next;
} CHILD;

enum sym_type {
    sym_begin, sym_end,
    sym_empty_2, sym_branch_2, sym_vert_2, sym_last_2,
    sym_single_3, sym_first_3
};
    
static char const * const sym_ascii[] = {
    "", "",
    "  ", "|-", "| ", "`-",
    "---", "-+-"
};
static char const * const sym_utf[] = {
    "", "",
    "  ", UTF_VR UTF_H, UTF_V " ", UTF_UR UTF_H,
    UTF_H UTF_H UTF_H, UTF_H UTF_HD UTF_H
};
static char const * const sym_vt100[] = {
    VT_BEG, VT_END,
    "  ", VT_VR VT_H, VT_V " ", VT_UR VT_H,
    VT_H VT_H VT_H, VT_H VT_HD VT_H
};

static char const * const * sym = sym_ascii;

static PROC *list = NULL, *vroot = NULL;
static int width[MAX_DEPTH],more[MAX_DEPTH];
int print_args = 0,compact = 1,user_change = 0,pids = 0,by_pid = 0,
  trunc = 1;
static int output_width = 0;
static int cur_x = 1;
static char last_char = 0;
static int dumped = 0; /* used by dump_by_user */
char const *progname;

static void out_char(char c)
{
    cur_x += (c & 0xc0) != 0x80; /* only count first UTF-8 char */
    if (cur_x <= output_width || !trunc) putchar(c);
    if (cur_x == output_width+1 && trunc)
	if (last_char || (c & 0x80)) putchar('+');
	else {
	    last_char = c;
	    cur_x--;
	    return;
	}
}


static void out_string(const char *str)
{
    while (*str) out_char(*str++);
}


static void out_sym(enum sym_type which)
{

    fputs(sym[sym_begin], stdout);
    out_string(sym[which]);
    fputs(sym[sym_end], stdout);
}


static int out_int(int x) /* non-negative integers only */
{
    int digits,div;

    if (x == 0) {
	out_char('0');
	return 1;
    }

    digits = 0;
    for (div = 1; x/div; div *= 10) digits++;
    for (div /= 10; div; div /= 10) out_char('0'+(x/div) % 10);
    return digits;
}


static void out_newline(void)
{
    if (last_char && cur_x == output_width) putchar(last_char);
    last_char = 0;
    putchar('\n');
    cur_x = 1;
}


static PROC *find_proc(pid_t pid)
{
    PROC *walk;

    for (walk = list; walk; walk = walk->next)
	if (walk->pid == pid) break;
    return walk;
}


static void add_child(PROC *parent,PROC *child)
{
    CHILD *new,**walk;
    int cmp;

    if (!(new = malloc(sizeof(CHILD))))
        errx(1, "malloc failed");
    new->child = child;
    for (walk = &parent->children; *walk; walk = &(*walk)->next)
	if (by_pid) {
	    if ((*walk)->child->pid > child->pid) break;
	}
	else if ((cmp = strcmp((*walk)->child->comm,child->comm)) > 0) break;
	    else if (!cmp && (*walk)->child->uid > child->uid) break;
    new->next = *walk;
    *walk = new;
}

/*
 * Turn a Linux-style null-separated argument list into an argv/argc-style
 * one.  The argv returned should be free()d when finished with.
 * We depend on there being a NUL at the end of the list.
 */

static void mkargv(int *argcp, char const ***argvp, char const *args, int size)
{
    int argc;
    char const **argv;
    int i;

    argc = 0;
    for (i = 0; i < size; i++)
	if (args[i] == '\0')
	    argc++;
    if ((argv = malloc(sizeof(char *) * argc)) == NULL)
	errx(1, "malloc failed");
    argv[0] = args;
    for (i = 1; i < argc; i++)
	argv[i] = strchr(argv[i-1], '\0') + 1;
    *argvp = argv; *argcp = argc;
}

int count_argv(char const * const *argv)
{
    int i;

    i = 0;
    if (argv == NULL) return 0;
    while (argv[i] != NULL)
	i++;
    return i;
}


/*
 * Add a process to the list.  comm and argv will be copied.
 * Linking processes into a tree happens later.
 */

void add_proc(const char *comm,pid_t pid,pid_t ppid,uid_t uid, int argc,
	      char const * const *argv)
{
    PROC *this;
    int i;

    if (find_proc(pid) != NULL) {
	warnx("Duplicate process (pid=%d)", pid);
	return;
    }
    if ((this = malloc(sizeof(PROC))) == NULL)
	errx(1, "malloc failed");
    if ((this->comm = strdup(comm)) == NULL)
	errx(1, "strdup failed");
    this->pid = pid;
    this->ppid = ppid;
    this->uid = uid;
    this->highlight = 0;
    this->children = NULL;
    this->parent = NULL;
    this->uid = uid;
    if (argv == NULL) {
	this->argc = 0;
	this->argv = NULL;
    } else {
	this->argc = argc;
	this->argv = malloc(sizeof(char *) * argc);
	if (this->argv == NULL)
	    errx(1, "malloc failed");
	for (i = 0; i < argc; i++)
	    if ((this->argv[i] = strdup(argv[i])) == NULL)
		errx(1, "strdup failed");
    }
    this->next = list;
    list = this;
}


void add_proc_args(const char *comm,pid_t pid,pid_t ppid,uid_t uid,
			  char const *args, int size)
{
    int argc;
    char const **argv;

    mkargv(&argc, &argv, args, size);
    add_proc(comm, pid, ppid, uid, argc, argv);
    free(argv);
}

/*
 * Tie a given process into the process tree
 */

static void attach_proc(PROC *this)
{

    if (this->pid == this->ppid ||
	(this->parent = find_proc(this->ppid)) == NULL )
	this->parent = vroot;
    add_child(this->parent,this);
}


static void knit_tree(void)
{
    PROC *this;

    if ((vroot = malloc(sizeof(PROC))) == NULL)
	errx(1, "malloc failed");
    memset(vroot, 0, sizeof(*vroot));
    for (this = list; this; this = this->next)
	attach_proc(this);
}


static int tree_equal(const PROC *a,const PROC *b)
{
    const CHILD *walk_a,*walk_b;

    if (strcmp(a->comm,b->comm)) return 0;
    if (user_change && a->uid != b->uid) return 0;
    for (walk_a = a->children, walk_b = b->children; walk_a && walk_b;
      walk_a = walk_a->next, walk_b = walk_b->next)
	if (!tree_equal(walk_a->child,walk_b->child)) return 0;
    return !(walk_a || walk_b);
}


static void dump_tree(PROC *current,int level,int rep,int leaf,int last,
  uid_t prev_uid,int closing)
{
    CHILD *walk,*next,**scan;
    const struct passwd *pw;
    int lvl,i,add,offset,len,swapped,info,count,comm_len,first;
    const char *here;
#ifdef HAVE_TERMCAP
    char tbuf[1024], *tbufptr;
    char *tmp;
#endif
#ifdef HAVE_TERMINFO
    char *tmp;
#endif
    char comm_tmp[5];

    if (!current) return;
    if (level >= MAX_DEPTH-1) {
	fprintf(stderr,"MAX_DEPTH not big enough.\n");
	exit(1);
    }
    if (!leaf)
	for (lvl = 0; lvl < level; lvl++) {
	    for (i = width[lvl]+1; i; i--) out_char(' ');
	    out_sym(lvl == level-1 ? last ? sym_last_2 : sym_branch_2 :
	       more[lvl+1] ? sym_vert_2 : sym_empty_2);
	}
    if (rep < 2) add = 0;
    else {
	add = out_int(rep)+2;
	out_string("*[");
    }
#ifdef HAVE_TERMCAP
    tbufptr = tbuf;
    if (current->highlight && (tmp = tgetstr("md",&tbufptr))) tputs(tmp,1,putchar);
#endif
#ifdef HAVE_TERMINFO
    if (current->highlight && (tmp = tigetstr("bold")) != (char *)-1)
	putp(tmp);
#endif
    if ((swapped = (print_args && current->argc <= 0))) out_char('(');
    comm_len = 0;
    for (here = current->comm; *here; here++)
	if (*here == '\\') {
	    out_string("\\\\");
	    comm_len += 2;
	}
	else if (*here > ' ' && *here <= '~') {
		out_char(*here);
		comm_len++;
	    }
	    else {
		sprintf(comm_tmp,"\\%03o",(unsigned char) *here);
		out_string(comm_tmp);
		comm_len += 4;
	    }
    offset = cur_x;
    info = pids || (user_change && prev_uid != current->uid);
    if (info) out_char(swapped ? ',' : '(');
    if (pids) (void) out_int(current->pid);
    if (user_change && prev_uid != current->uid) {
	if (pids) out_char(',');
	if ((pw = getpwuid(current->uid))) out_string(pw->pw_name);
	else (void) out_int(current->uid);
    }
    if (info || swapped) out_char(')');
#ifdef HAVE_TERMCAP
    tbufptr = tbuf;
    if (current->highlight && (tmp = tgetstr("me",&tbufptr)))
        tputs(tmp,1,putchar);
#endif
#ifdef HAVE_TERMINFO
    if (current->highlight && (tmp = tigetstr("sgr0")) != (char *)-1)
	putp(tmp);
#endif
    if (print_args) {
	for (i = 1; i < current->argc; i++) {
	    out_char(' ');
	    len = 0;
	    for (here = current->argv[i]; *here; here++)
		len += *here > ' ' && *here <= '~' ? 1 : 4;
	    if (!trunc || cur_x+len <= output_width-(i == current->argc-1 ? 0 : 4))
		for (here = current->argv[i]; *here; here++)
		    if (*here > ' ' && *here <= '~') out_char(*here);
		    else {
			sprintf(comm_tmp,"\\%03o",(unsigned char) *here);
			out_string(comm_tmp);
		    }
	    else {
		out_string("...");
		break;
	    }
	}
    }
    if (print_args || !current->children) {
	while (closing--) out_char(']');
	out_newline();
	if (print_args) {
	    more[level] = !last;
	    width[level] = swapped+(comm_len > 1 ? 0 : -1);
	    for (walk = current->children; walk; walk = walk->next)
		dump_tree(walk->child,level+1,1,0,!walk->next,current->uid,0);
	}
    }
    else {
	more[level] = !last;
	width[level] = comm_len+cur_x-offset+add;
	if (cur_x >= output_width && trunc) {
	    out_sym(sym_first_3);
	    out_string("+");
	    out_newline();
	}
	else {
	    first = 1;
	    for (walk = current->children; walk; walk = next) {
		count = 0;
		next = walk->next;
		if (compact) {
		    scan = &walk->next;
		    while (*scan)
			if (!tree_equal(walk->child,(*scan)->child)) 
			    scan = &(*scan)->next;
			else {
			    if (next == *scan) next = (*scan)->next;
			    count++;
			    *scan = (*scan)->next;
			}
		}
		if (first) {
		    out_sym(next ? sym_first_3 : sym_single_3);
		    first = 0;
		}
		dump_tree(walk->child,level+1,count+1,walk == current->children,
		  !next,current->uid,closing+(count ? 1 : 0));
	    }
	}
    }
}


static void dump_by_user(PROC *current,uid_t uid)
{
    const CHILD *walk;

    if (current->uid == uid && current != vroot) {
	if (dumped) putchar('\n');
	dump_tree(current,0,1,1,1,uid,0);
	dumped = 1;
	return;
    }
    for (walk = current->children; walk; walk = walk->next)
	dump_by_user(walk->child,uid);
}

void read_proc_generic(char const *proc_base,
		       void (*read_proc_entry)(char const *))
{
    DIR *dir;
    struct dirent *de;

    dir = opendir(proc_base);
    if (dir == NULL) return;
    while ((de = readdir(dir)) != NULL)
	if (de->d_name[strspn(de->d_name, "0123456789")] == '\0') {
	    if (chdir(proc_base) != 0)
		warn("%s", proc_base);
	    (*read_proc_entry)(de->d_name);
	}
    closedir(dir);
}

static void read_proc_entry_linux(char const *name)
{
    FILE *file;
    int fd;
    pid_t pid, ppid;
    char comm[COMM_LEN+1]; /* FIXME */
    struct stat st;
    char *buffer;
    ssize_t size;

    if (!print_args) buffer = NULL;
    else if (!(buffer = malloc((size_t) (output_width+1))))
	errx(1, "malloc failed");
    pid = atoi(name);
    if (chdir(name) == 0 && (file = fopen("stat", "r")) != NULL) {
	if (fscanf(file,"%*d (%[^)]) %*c %d", comm, &ppid) == 2) {
	    if (stat(".",&st) == 0) {
		if (!print_args)
		    add_proc(comm,pid,ppid,st.st_uid,0,NULL);
		else {
		    if ((fd = open("cmdline",O_RDONLY)) >= 0) {
			size = read(fd,buffer,(size_t) output_width);
			if (size >= 0) {
			    if (size) buffer[size++] = 0;
			    add_proc_args(comm,pid,ppid,st.st_uid,buffer,size);
			}
			close(fd);
		    }
		}
	    }
	}
	fclose(file);
    }
    if (print_args) free(buffer);
}

static void read_proc_entry_bsd(char const *name)
{
    FILE *file;
    int fd;
    pid_t pid, ppid;
    uid_t uid;
    char comm[COMM_LEN+1]; /* FIXME */
    char *buffer;
    ssize_t size;

    if (!print_args) buffer = NULL;
    else if (!(buffer = malloc((size_t) (output_width+1))))
	errx(1, "malloc failed");
    /* Must be checked after Solaris and Linux.  They both
       have their own idea of what /proc/%d/status is for.  */
    if (chdir(name) == 0 && (file = fopen("status", "r")) != NULL) {
	if (fscanf(file, "%s %d %d %*d %*d %*d,%*d %*s "
		   "%*d,%*d %*d,%*d %*d,%*d %*s %d",
		   comm, &pid, &ppid, &uid) == 4) {
	    if (!print_args)
		add_proc(comm,pid,ppid,uid,0,NULL);
	    else {
		if ((fd = open("cmdline",O_RDONLY)) < 0) {
		    if (errno == ENOENT)
			/* Old versions don't have it */
			add_proc(comm,pid,ppid,uid,0,NULL);
		} else {
		    size = read(fd,buffer,(size_t) output_width);
		    if (size >= 0) {
			if (size) buffer[size++] = 0;
			add_proc_args(comm,pid,ppid,uid,buffer,size);
		    }
		    close(fd);
		}
	    }
	}
	fclose(file);
    }
    if (print_args) free(buffer);
}

static void read_proc_linux(void)
{
    struct stat st;

    if (stat("/proc/self/stat", &st) == 0)
	read_proc_generic("/proc", &read_proc_entry_linux);
}    

static void read_proc_bsd(void)
{
    struct stat st;

    if (stat("/proc/curproc/status", &st) == 0)
	read_proc_generic("/proc", &read_proc_entry_bsd);
}

static void usage(void)
{
    fprintf(stderr,"usage: pstree [ -a ] [ -c ] [ -h | -H pid ] [ -l ] [ -n ] "
      "[ -p ] [ -u ]\n%14s[ -G | -U ] [ pid | user]\n","");
    fprintf(stderr,"       pstree -V\n\n");
    fprintf(stderr,"    -a     show command line arguments\n");
    fprintf(stderr,"    -c     don't compact identical subtrees\n");
#if defined(HAVE_TERMCAP) || defined(HAVE_TERMINFO)
    fprintf(stderr,"    -h     highlight current process and its ancestors\n");
    fprintf(stderr,"    -H pid highlight process \"pid\" and its ancestors\n");
#endif
    fprintf(stderr,"    -G     use VT100 line drawing characters\n");
    fprintf(stderr,"    -l     don't truncate long lines\n");
    fprintf(stderr,"    -n     sort output by PID\n");
    fprintf(stderr,"    -p     show PIDs; implies -c\n");
    fprintf(stderr,"    -u     show uid transitions\n");
    fprintf(stderr,"    -U     use UTF-8 (Unicode) line drawing characters\n");
    fprintf(stderr,"    -V     display version information\n");
    fprintf(stderr,"    pid    start at pid, default 1 (init)\n");
    fprintf(stderr,"    user   show only trees rooted at processes of that "
      "user\n\n");
    exit(1);
}


int main(int argc,char **argv)
{
    PROC *current;
    const struct passwd *pw;
    pid_t pid,highlight;
    int highlightp;
#ifdef HAVE_TERMCAP
    char termcap_area[1024];
#endif
#ifdef HAVE_TERMINFO
    int err;
#endif
#ifdef TIOCGWINSZ
    struct winsize winsz;
#endif
    int c;

    progname = argv[0];
    pid = 1;
    highlight = 0;
    highlightp = 0;
    pw = NULL;
    while ((c = getopt(argc,argv,"acGhH:npluUVw")) != EOF)
	switch (c) {
	    case 'a':
		print_args = 1;
		break;
	    case 'c':
		compact = 0;
		break;
	    case 'G':
		if (sym != sym_ascii) usage();
		sym = sym_vt100;
		break;
	    case 'h':
		if (highlightp) usage();
		highlight = getpid();
		highlightp = 1;
		break;
	    case 'H':
		if (highlightp)
		    usage();
		errno = 0;
		highlight = strtol(optarg, NULL, 0);
		if (errno != 0)
		    usage();
		highlightp = 1;
		break;
	    case 'l':
		trunc = 0;
		break;
	    case 'n':
		by_pid = 1;
		break;
	    case 'p':
		pids = 1;
		compact = 0;
		break;
	    case 'u':
		user_change = 1;
		break;
	    case 'U':
		if (sym != sym_ascii) usage();
		sym = sym_utf;
		break;
	    case 'V':
#ifdef PSMISC_VERSION
		fprintf(stderr,"pstree from psmisc version " PSMISC_VERSION 
			"\n");
#else
#ifdef PSTREE_VERSION
		fprintf(stderr, "pstree version " PSTREE_VERSION "\n");
#else
		fprintf(stderr, "pstree (unknown version)\n");
#endif
#endif
		return 0;
	    default:
		usage();
	}
#ifdef HAVE_TERMCAP
    if (!getenv("TERM")) {
	if (highlightp)
	    warnx("TERM is not set");
	highlight = 0;
    } else {
	if (tgetent(termcap_area,getenv("TERM")) <= 0) {
	    if (highlightp)
		warnx("can't get terminal capabilities");
	    highlightp = 0;
	}
    }
#endif
#ifdef HAVE_TERMINFO
    setupterm(NULL, fileno(stdout), &err);
    switch (err) {
        case -1:
	    if (highlightp)
		warnx("Terminfo database not found");
	    highlightp = 0;
	    break;
	case 0:
	    if (highlightp)
		warnx("Terminfo entry not found");
	    highlightp = 0;
	    break;
	case 1:
	    output_width = COLS;
	    break;
    }
#endif
    if (output_width == 0 && getenv("COLUMNS") != NULL)
	output_width = atoi(getenv("COLUMNS"));
#ifdef TIOCGWINSZ
    if (output_width == 0 && ioctl(1,TIOCGWINSZ,&winsz) >= 0)
    	output_width = winsz.ws_col;
#endif
    if (output_width == 0)
	output_width = 132;
    
    if (optind == argc-1)
	if (isdigit(*argv[optind])) {
	    errno = 0;
	    pid = strtol(argv[optind++], NULL, 0);
	    if (errno != 0)
		usage();
	} else if (!(pw = getpwnam(argv[optind++]))) {
	    fprintf(stderr,"No such user name: %s\n",argv[optind-1]);
	    return 1;
	}
    if (optind != argc) usage();
#ifdef HAVE_LIBPS
    if (list == NULL) read_libps();
#endif
#ifdef HAVE_LIBPROC
    if (list == NULL) read_libproc();
#endif
#ifdef HAVE_STRUCT_PSINFO
    if (list == NULL) read_proc_solaris();
#endif
    if (list == NULL) read_proc_linux();
    if (list == NULL) read_proc_bsd();
#ifdef HAVE_STRUCT_PRPSINFO
    if (list == NULL) read_proc_irix();
    if (list == NULL) read_proc_svr4();
#endif
    if (list == NULL) errx(1, "Could't get process information");
    knit_tree();
    if (highlightp)
	for (current = find_proc(highlight); current;
	     current = current->parent)
	    current->highlight = 1;
    if (!pw)
	dump_tree(find_proc(pid),0,1,1,1,0,0);
    else {
	dump_by_user(vroot,pw->pw_uid);
	if (!dumped) {
	    fprintf(stderr,"No processes found.\n");
	    return 1;
	}
    }
    return 0;
}

/*
 * Local variables:
 * c-file-style: "bsd"
 * End:
 */
